
from .speech_transcribe_whisper import (
    speech_transcription_whisper,
)

from .speech_separation_nlabels import (
    speaker_separation_nolabels,
)

from .speech_separation_labels import (
    speaker_separation_labels,
)

from .speech_transcribe_cloud import (
    speech_transcription_aws,
)

from .speech_transcribe_vosk import (
    speech_transcription_vosk,
)

from .willisdiarize_aws import (
    diarization_correction_aws,
)

from .willisdiarize import (
    diarization_correction
)

from .commons import (
    to_audio,
)


__all__ = ["speech_transcription_whisper", "speaker_separation_nolabels", "speaker_separation_labels", "speech_transcription_aws", "speech_transcription_vosk", "diarization_correction_aws", "diarization_correction", "to_audio"]



from .common import (
    to_audio, from_audio, get_config
)

from .common_separation import (
    transcribe_response_to_dataframe, whisperx_to_dataframe, vosk_to_dataframe,
    volume_normalization
)

__all__ = ["to_audio", "get_config", "from_audio", "transcribe_response_to_dataframe", "whisperx_to_dataframe", "vosk_to_dataframe", "volume_normalization"]



# author:    Vijay Yadav
# website:   http://www.bklynhlth.com

# import the required packages
import json
import os
from pydub import AudioSegment
import numpy as np
import logging

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

def make_dir(dir_name):
    """
    ------------------------------------------------------------------------------------------------------

    Creates a directory if it doesn't already exist.

    Parameters:
    ...........
    dir_name : str
        The path to the directory

    Returns:
    ...........
    None

    ------------------------------------------------------------------------------------------------------
    """
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

def to_audio(filepath, speaker_dict, output_dir, cleaned=False):
    """
    ------------------------------------------------------------------------------------------------------

    Save a numpy signal into multiple speaker-specific audio files.

    Parameters:
    ----------
        filepath : str
            The path to the input audio file.
        speaker_dict : dict
            A dictionary containing speaker labels as keys and corresponding segments (NumPy arrays) as values.
        output_dir : str
            The directory where the output audio files will be saved.
        cleaned : bool
            A flag indicating whether the audio signal has been cleaned.

    ------------------------------------------------------------------------------------------------------
    """
    make_dir(output_dir)
    for key, value in speaker_dict.items():
        file_name, _ = os.path.splitext(os.path.basename(filepath))

        audio_signal = AudioSegment.from_file(file = filepath, format = "wav")
        spk_signal = AudioSegment(value.tobytes(), frame_rate=44100 if cleaned else audio_signal.frame_rate,
                                  sample_width=audio_signal.sample_width, channels=audio_signal.channels)

        output_file = os.path.join(output_dir, file_name + '_' + key + '.wav')
        spk_signal.export(output_file, format="wav")

def from_audio(audio_dir):
    """
    ------------------------------------------------------------------------------------------------------
    
    Load multiple speaker-specific audio files and reconstruct a dictionary with speaker labels as keys 
    and corresponding segments (NumPy arrays) as values.

    Parameters:
    ----------
        audio_dir : str
            The directory containing the speaker-specific audio files.

    Returns:
    ----------
        speaker_dict : dict
            A dictionary with speaker labels as keys and corresponding audio data (NumPy arrays) as values.

    ------------------------------------------------------------------------------------------------------
    """
    speaker_dict = {}
    
    # Iterate over the files in the directory
    for file_name in os.listdir(audio_dir):
        if file_name.endswith('.wav'):
            try:
                # Extract speaker key from the file name (assuming the format 'filename_speaker.wav')
                speaker_key = file_name.split('_')[-1].replace('.wav', '')
                
                # Load the audio file
                audio_file_path = os.path.join(audio_dir, file_name)
                audio = AudioSegment.from_file(audio_file_path, format="wav")
                
                # Convert audio to raw bytes and then to a numpy array
                audio_array = np.array(audio.get_array_of_samples())
                
                # Store in the dictionary with speaker key
                speaker_dict[speaker_key] = audio_array
            except Exception as e:
                logger.info(f"Error while converting from audio to numpy array file: {file_name}, error: {e}")
    
    return speaker_dict

def get_config(filepath, json_file):
    """
    ------------------------------------------------------------------------------------------------------

    This function reads the configuration file containing the column names for the output dataframes,
    and returns the contents of the file as a dictionary.

    Parameters:
    ...........
    filepath : str
        The path to the configuration file.
    json_file : str
        The name of the configuration file.

    Returns:
    ...........
    measures: A dictionary containing the names of the columns in the output dataframes.

    ------------------------------------------------------------------------------------------------------
    """
    dir_name = os.path.dirname(filepath)
    measure_path = os.path.abspath(os.path.join(dir_name, f"config/{json_file}"))

    file = open(measure_path)
    measures = json.load(file)
    return measures



# author:    Vijay Yadav
# website:   http://www.brooklyn.health

# import the required packages

import boto3
import urllib
import time
import random

import json
import logging
from . import separation_util as sutil

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

def replace_speaker_labels(data, check_labels, speaker_labels):
    """
    ------------------------------------------------------------------------------------------------------

    Recursively replaces speaker labels in a nested dictionary or list.

    Parameters:
    ...........
    data : dict or list
        The nested dictionary or list containing speaker labels.
    check_labels: list
        Check on input speaker labels
    speaker_labels: list
        Expected speaker labels

    Returns:
    ...........
    data : dict or list
        The modified nested dictionary or list with replaced speaker labels.

    ------------------------------------------------------------------------------------------------------
    """

    if isinstance(data, dict):
        for key, value in data.items():

            if key == 'speaker_label':
                if value == check_labels[0]:
                    data[key] = speaker_labels[0]

                elif value == check_labels[1]:
                    data[key] = speaker_labels[1]

                else:
                    data[key] = value
            else:
                replace_speaker_labels(value, check_labels, speaker_labels)
    elif isinstance(data, list):
        for item in data:
            replace_speaker_labels(item, check_labels, speaker_labels)

    return data

def filter_labels_aws(data):
    """
    ------------------------------------------------------------------------------------------------------

    replaces speaker labels in AWS JSON.

    Parameters:
    ...........
    data : JSON
        The JSON containing speaker labels.

    Returns:
    ...........
    data : JSON
        The modified JSON with replaced speaker labels.

    ------------------------------------------------------------------------------------------------------
    """
    if 'results' in data:
        speaker_labels = data['results'].get('speaker_labels', {})
        segments = speaker_labels.get('segments', [])

        for segment in segments:
            seg_speaker_label = segment.get('speaker_label', '')
            
            if 'spk_' in seg_speaker_label:
                segment['speaker_label'] = seg_speaker_label.replace("spk_", "speaker")
            
            seg_items = segment.get('items', [])
            for seg_item in seg_items:
                
                seg_item_speaker_label = seg_item.get('speaker_label', '')
                if 'spk_' in seg_item_speaker_label:
                    
                    seg_item['speaker_label'] = seg_item_speaker_label.replace("spk_", "speaker")
        items = data['results'].get('items', [])
        
        for item in items:
            item_speaker_label = item.get('speaker_label', '')
            
            if 'spk_' in item_speaker_label:
                item['speaker_label'] = item_speaker_label.replace("spk_", "speaker")

    return data

def filter_labels_whisper(data):
    """
    ------------------------------------------------------------------------------------------------------

    replaces speaker labels in Whisper JSON.

    Parameters:
    ...........
    data : JSON
        The JSON containing speaker labels.

    Returns:
    ...........
    data : JSON
        The modified JSON with replaced speaker labels.

    ------------------------------------------------------------------------------------------------------
    """
    for segment in data.get('segments', []):
        current_speaker = segment.get('speaker', '')

        if 'SPEAKER_0' in current_speaker:
            segment["speaker"] = current_speaker.replace("SPEAKER_0", "speaker")

        for word in segment["words"]:
            word_speaker = word.get('speaker', '')
            
            if 'SPEAKER_0' in word_speaker:
                word["speaker"] = word_speaker.replace("SPEAKER_0", "speaker")

    for word_segment in data.get('word_segments', []):
        word_seg_speaker = word_segment.get('speaker', '')
        
        if 'SPEAKER_0' in word_seg_speaker: 
            word_segment["speaker"] = word_seg_speaker.replace("SPEAKER_0", "speaker")

    return data

def extract_content(data):
    """
    ------------------------------------------------------------------------------------------------------

    Extracts content from a nested dictionary and returns it in a speaker-based dictionary.

    Parameters:
    ...........
    data: json
        Speech transcription response

    Returns:
    ...........
    content_dict : dict
        A dictionary where the keys are speaker labels and the values are the corresponding content spoken by each

    ------------------------------------------------------------------------------------------------------
    """
    content_dict = {}

    if 'results' in data:           # Check if 'results' key is present
        item_data = data['results']['items']

        item_spk_0 = [item for item in item_data if item.get('speaker_label', '') == 'speaker0']
        item_spk_1 = [item for item in item_data if item.get('speaker_label', '') == 'speaker1']

        spk_0_text = [item['alternatives'][0]['content'] for item in item_spk_0 if 'alternatives' in item]
        spk_1_text = [item['alternatives'][0]['content'] for item in item_spk_1 if 'alternatives' in item]

        content_dict['speaker0'] = " ".join(spk_0_text)
        content_dict['speaker1'] = " ".join(spk_1_text)

    return content_dict

def exponential_backoff_decorator(max_retries, base_delay):
    def decorator(func):
        def wrapper(*args, **kwargs):
            retries = 0
            while retries < max_retries:
                try:
                    result_func = func(*args, **kwargs)
                    return result_func
                except Exception as e:
                    retries += 1
                    delay = base_delay * 2**retries + random.uniform(0, 1)
                    logging.error(f"Retrying in {delay:.2f} seconds...")
                    time.sleep(delay)
            raise Exception("Max retries reached, operation failed.")

        return wrapper

    return decorator

@exponential_backoff_decorator(max_retries=3, base_delay=90)
def get_clinical_labels(scale, measures, content_dict, json_response, context_model):
    """
    ------------------------------------------------------------------------------------------------------

    Replaces speaker labels in a JSON response based on clinical measures.

    Parameters:
    ...........
    scale : str
        Clinical scale
    measures : object
        A configuration object.
    content_dict: dict
        A dictionary containing speaker-based content.
    json_response: json
        Speech transcription response
    context_model: str
        A string containing the name of the embedding model to be used.

    Returns:
    ...........
    json_response : json
        The modified JSON response with replaced speaker labels.

    ------------------------------------------------------------------------------------------------------
    """
    #Check if content is available for all the speaker
    if len(content_dict) <2:
        return json_response

    model, tokenizer = sutil.get_embedding_model(context_model)

    score_string = scale.lower()+'_string'
    spk1_score = sutil.match_transcript(measures[score_string], content_dict['speaker0'], model, tokenizer)
    spk2_score = sutil.match_transcript(measures[score_string], content_dict['speaker1'], model, tokenizer)

    if spk1_score > spk2_score:
        json_response = replace_speaker_labels(json_response, ['speaker0', 'speaker1'], ['clinician', 'participant'])

    else:
        json_response = replace_speaker_labels(json_response, ['speaker0', 'speaker1'], ['participant', 'clinician'])
    
    return json_response

def get_job_status(transcribe, input_param):
    """
    ------------------------------------------------------------------------------------------------------

    Get transcriptopn job status

    Parameters:
    ...........
    transcribe : object
        Transcription client object
    input_param: dict
        A dictionary containing input parameters with their corresponding values.

    Returns:
    ...........
    status : str
        Transcription job status

    ------------------------------------------------------------------------------------------------------
    """
    status = json.loads("{}")
    while True:
        try:

            sleep_time = int(random.uniform(5, 10))
            status = transcribe.get_transcription_job(TranscriptionJobName=input_param['job_name'])

            if status['TranscriptionJob']['TranscriptionJobStatus'] in ['COMPLETED', 'FAILED']:
                break

            time.sleep(sleep_time)
        except Exception as e:
            logger.info(f"Get Transcription job failed with exception: {e}")

    return status

def filter_transcript_response(status, input_param):
    """
    ------------------------------------------------------------------------------------------------------

    Filter transcriptopn job response

    Parameters:
    ...........
    status : object
        Transcription job response
    input_param: dict
        A dictionary containing input parameters with their corresponding values.

    Returns:
    ...........
    tuple : Two strings, the first string is the response object in JSON format, and the second string is
    the transcript of the audio file.

    ------------------------------------------------------------------------------------------------------
    """
    read_data = urllib.request.urlopen(status['TranscriptionJob']['Transcript']['TranscriptFileUri'])
    response = json.loads(read_data.read().decode('utf-8'))

    transcript = response['results']['transcripts'][0]['transcript']
    if input_param['speaker_labels'] == True:#replace speaker labels with standard names

        response = filter_labels_aws(response)
    return response, transcript

def transcribe_audio(s3uri, input_param):
    """
    ------------------------------------------------------------------------------------------------------

    Transcribe an audio file using AWS Transcribe.

    Parameters:
    ...........
    s3uri : str
        The S3 uri for the recording to be transcribed.
    input_param: dict
        A dictionary containing input parameters with their corresponding values.

    Raises:
    ...........
    Exception: If the transcription job fails.

    Returns:
    ...........
    tuple : Two strings, the first string is the response object in JSON format, and the second string is
    the transcript of the audio file.

    ------------------------------------------------------------------------------------------------------
    """
    response = json.dumps({})
    settings = {}
    transcript = ""

    try:
        if input_param['access_key'] and input_param['secret_key']:
            transcribe = boto3.client('transcribe', region_name = input_param['region'], 
                                      aws_access_key_id = input_param['access_key'], 
                                      aws_secret_access_key = input_param['secret_key'])
        else:
            transcribe = boto3.client('transcribe', region_name = input_param['region'])

        if input_param['speaker_labels'] == True and input_param['max_speakers']>=2:
            settings = {'ShowSpeakerLabels': input_param['speaker_labels'], 'MaxSpeakerLabels': input_param['max_speakers']}

        transcribe.start_transcription_job(
            TranscriptionJobName=input_param['job_name'],
            Media={'MediaFileUri': s3uri},
            LanguageCode=input_param['language'],
            Settings=settings)

        status = get_job_status(transcribe, input_param)
        if status['TranscriptionJob']['TranscriptionJobStatus'] == 'COMPLETED':
            response, transcript = filter_transcript_response(status, input_param)

    except Exception as e:
        logger.info(f"AWS Transcription job failed with file: {s3uri} exception: {e}")

    finally:
        return response, transcript

def get_whisperx_content(data):
    """
    ------------------------------------------------------------------------------------------------------

    Extracts content from a nested dictionary and returns it in a speaker-based dictionary.

    Parameters:
    ...........
    data: json
        Speech transcription response

    Returns:
    ...........
    content_dict : dict
        A dictionary where the keys are speaker labels and the values are the corresponding content spoken by each

    ------------------------------------------------------------------------------------------------------
    """
    content_dict = {}

    if 'segments' in data:           # Check if 'segments' key is present
        item_data = data['segments']
        
        spk_0_text = [item['text'] for item in item_data if item.get('speaker', '') == 'speaker0']
        spk_1_text = [item['text'] for item in item_data if item.get('speaker', '') == 'speaker1']
        
        spk0_check = any(not char.isspace() for char in spk_0_text)
        spk1_check = any(not char.isspace() for char in spk_1_text)

        if spk0_check and spk1_check:
            content_dict['speaker0'] = " ".join(spk_0_text)
            content_dict['speaker1'] = " ".join(spk_1_text)

    return content_dict

def replace_whisperx_speaker_labels(json_data, check_label, replace_label):
    """
    ------------------------------------------------------------------------------------------------------

    Replaces speaker labels in json response.

    Parameters:
    ...........
    data : json
        The json containing speaker labels.
    check_labels: list
        Check on input speaker labels
    speaker_labels: list
        Expected speaker labels

    Returns:
    ...........
    data : json
        The modified json with replaced speaker labels.

    ------------------------------------------------------------------------------------------------------
    """
    updated_data = json_data.copy()
    if 'segments' in updated_data:
        
        for segment in updated_data['segments']:
            if 'speaker' not in segment:
                continue
                
            if segment['speaker'] == check_label[0]:
                segment['speaker'] = replace_label[0]
    
            elif segment['speaker'] == check_label[1]:
                segment['speaker'] = replace_label[1]

    return updated_data

def get_whisperx_clinical_labels(scale, measures, content_dict, json_response, context_model):
    """
    ------------------------------------------------------------------------------------------------------

    Replaces speaker labels in a JSON response based on clinical measures.

    Parameters:
    ...........
    scale : str
        Clinical scale
    measures : object
        A configuration object.
    content_dict: dict
        A dictionary containing speaker-based content.
    json_response: json
        Speech transcription response
    context_model: str
        A string containing the name of the embedding model to be used.

    Returns:
    ...........
    json_response : json
        The modified JSON response with replaced speaker labels.

    ------------------------------------------------------------------------------------------------------
    """
    #Check if content is available for all the speaker
    if len(content_dict) <2:
        return json_response

    model, tokenizer = sutil.get_embedding_model(context_model)

    score_string = scale.lower()+'_string'
    spk1_score = sutil.match_transcript(measures[score_string], content_dict['speaker0'], model, tokenizer)
    spk2_score = sutil.match_transcript(measures[score_string], content_dict['speaker1'], model, tokenizer)
    
    if spk1_score > spk2_score:
        json_response = replace_whisperx_speaker_labels(json_response, ['speaker0', 'speaker1'], ['clinician', 'participant'])

    else:
        json_response = replace_whisperx_speaker_labels(json_response, ['speaker0', 'speaker1'], ['participant', 'clinician'])

    return json_response



# author:    Vijay Yadav
# website:   http://www.brooklyn.health

# import the required packages
import os
import json
import logging
from .util import transcribe_util as tutil
from .willisdiarize_aws import diarization_correction_aws

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

def get_config():
    """
    ------------------------------------------------------------------------------------------------------

    Load the configuration settings for the speech transcription.

    Parameters:
    ...........
    None

    Returns:
    ...........
    measures : dict
        A dictionary containing the configuration settings.

    ------------------------------------------------------------------------------------------------------
    """
    #Loading json config
    dir_name = os.path.dirname(os.path.abspath(__file__))
    measure_path = os.path.abspath(os.path.join(dir_name, 'config/speech.json'))

    file = open(measure_path)
    measures = json.load(file)
    return measures

def read_kwargs(kwargs):
    """
    ------------------------------------------------------------------------------------------------------

    Reads keyword arguments and returns a dictionary containing input parameters.

    Parameters:
    ...........
    kwargs : dict
        Keyword arguments to be processed.

    Returns:
    ...........
    input_param: dict A dictionary containing input parameters with their corresponding values.

    ------------------------------------------------------------------------------------------------------
    """
    input_param = {}
    input_param['language'] = kwargs.get('language', 'en-US')
    input_param['region'] = kwargs.get('region', 'us-east-1')

    input_param['job_name'] = kwargs.get('job_name', 'transcribe_job_01')
    input_param['speaker_labels'] = kwargs.get('speaker_labels', False)
    input_param['max_speakers'] = kwargs.get('max_speakers', 2)

    input_param['context'] = kwargs.get('context', '')
    input_param['context_model'] = kwargs.get('context_model', 'sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2')
    input_param['access_key'] = kwargs.get('access_key', '')
    input_param['secret_key'] = kwargs.get('secret_key', '')

    input_param['willisdiarize_endpoint'] = kwargs.get('willisdiarize_endpoint', '')
    input_param['willisdiarize_parallel'] = kwargs.get('willisdiarize_parallel', 1)

    return input_param

def speech_transcription_aws(s3_uri, **kwargs):
    """
    ------------------------------------------------------------------------------------------------------

    Speech transcription function that transcribes an audio file using Amazon Transcribe.

    Parameters:
    ...........
    s3_uri : str
        The S3 uri for the recording to be transcribed.
    kwargs: Object
        language : str, optional
            The language of the audio file (e.g. 'en-US', 'en-IN'). Default is 'en-US'.
        region : str, optional
            The AWS region to use (e.g. 'us-east-1'). Only applicable if model is 'aws'. Default is 'us-east-1'.
        job_name : str, optional
            The name of the transcription job. Only applicable if model is 'aws'. Default is 'transcribe_job_01'.
        access_key : str, optional
            AWS access key
        secret_key : str, optional
            AWS secret key
        speaker_labels : boolean, optional
            Show speaker labels
        max_speakers : int, optional
            Max number of speakers
        context : str, optional
            scale to use for slicing the separated audio files, if any.
        context_model : str, optional
            model to use for speaker identification, if context is provided.
        willisdiarize_endpoint : str, optional
            The SageMaker endpoint for the Willisdiarize API.
        willisdiarize_parallel : int, optional
            Whether to use parallel processing for Willisdiarize API.
            
    Returns:
    ...........
    json_response : JSON Object
        A transcription response object in JSON format
    transcript : str
        The transcript of the recording.
    willisdiarize_status : bool
        The status of the Willisdiarize API.

    ------------------------------------------------------------------------------------------------------
    """
    input_param = read_kwargs(kwargs)
    measures = get_config()
    json_response, transcript = tutil.transcribe_audio(s3_uri, input_param)
    willisdiarize_status = False

    present_labels = [x['speaker_label'] for x in json_response['results']['items'] if 'speaker_label' in x]
    if len(set(present_labels)) != 2:
        if input_param['speaker_labels'] == True and input_param['max_speakers'] == 1:
            for item in json_response['results']['items']:
                item['speaker_label'] = 'speaker_0'
        return json_response, transcript, willisdiarize_status

    if input_param['language'].lower()[:2] == 'en' and input_param['willisdiarize_endpoint'].lower() not in ['', 'none']:
        json_response, willisdiarize_status = diarization_correction_aws(
            json_response, input_param['willisdiarize_endpoint'],
            parallel_processing=input_param['willisdiarize_parallel'],
            region=input_param['region'], access_key=input_param['access_key'],
            secret_key=input_param['secret_key']
        )

    if input_param['speaker_labels'] == True and input_param['context'].lower() in measures['scale'].split(',') and input_param['context_model'] in measures['embedding_models']:
        content_dict = tutil.extract_content(json_response)
        json_response = tutil.get_clinical_labels(input_param['context'], measures, content_dict, json_response, input_param['context_model'])

    return json_response, transcript, willisdiarize_status



# author:    Georgios Efstathiadis
# website:   http://www.bklynhlth.com

# import the required packages
from transformers import AutoTokenizer, AutoModelForCausalLM
import torch

from .diarization_utils import (
    preprocess_str, apply_formatting
)


def process_chunk_hf(args):
    """
    ------------------------------------------------------------------------------------------------------

    This function processes a diarized text chunk using the HuggingFace model;
     it is used for parallel processing.

    Parameters:
    ...........
    args: tuple
        Tuple of arguments.
        idx: int, chunk index.
        prompt: str, diarized text chunk.
        model: HuggingFace model.
        tokenizer: HuggingFace tokenizer.

    Returns:
    ...........
    tuple: Tuple of chunk index and model output.

    ------------------------------------------------------------------------------------------------------
    """
    idx, prompt, model, tokenizer = args

    input_data = apply_formatting(preprocess_str(prompt))

    # Tokenize the input data
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    inputs = tokenizer(input_data["inputs"], return_tensors="pt")
    inputs = {key: value.to(device) for key, value in inputs.items()}

    # Generate the output
    output = model.generate(
        **inputs,
        max_new_tokens=2048,
        stop_strings=["</s>", "###"],
        tokenizer=tokenizer,
        return_dict_in_generate=True
    )

    # Decode the output
    result = tokenizer.decode(output["sequences"][0], skip_special_tokens=True)

    return idx, result


def call_diarization_hf(prompts, model_name, input_param):
    """
    ------------------------------------------------------------------------------------------------------

    This function calls the speaker diarization model.

    Parameters:
    ...........
    prompts: dict
        Dictionary of diarized text chunks.
        key: chunk index, value: diarized text chunk.
    model_name: str
        The name of the model used for transcription in HuggingFace.
    input_param: dict
        Additional arguments for the API call.

    Returns:
    ...........
    results: dict
        Dictionary of model outputs.
        key: chunk index, value: model output.

    ------------------------------------------------------------------------------------------------------
    """
    tokenizer = AutoTokenizer.from_pretrained(model_name, token=input_param['huggingface_token'])
    if torch.cuda.is_available():
        model = AutoModelForCausalLM.from_pretrained(model_name, token=input_param['huggingface_token'], device_map="cuda:0")
    else:
        raise ValueError("CUDA is not available.")

    results = {}
    for idx in sorted(prompts.keys()):
        results[idx] = process_chunk_hf((idx, prompts[idx], model, tokenizer))[1]

    return results



# author:    Vijay Yadav
# website:   http://www.bklynhlth.com

# import the required packages
import os
import json
import logging

from pyannote.audio import Pipeline
from .util import separation_util as sutil
from pydub import AudioSegment
import pandas as pd

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()
logging.getLogger('torch.distributed.nn.jit.instantiator').setLevel(logging.WARNING)

def get_config():
    """
    ------------------------------------------------------------------------------------------------------

    This function loads and returns a dictionary containing the configuration settings for speech separation
    from a JSON file.

    Parameters:
    ...........
    None

    Returns:
    ...........
    measures : dict
        A dictionary containing the configuration settings for speech separation.

    ------------------------------------------------------------------------------------------------------
    """
    #Loading json config
    dir_name = os.path.dirname(os.path.abspath(__file__))
    measure_path = os.path.abspath(os.path.join(dir_name, 'config/speech.json'))

    file = open(measure_path)
    measures = json.load(file)
    return measures

def run_pyannote(file_path, hf_token):
    """
    ------------------------------------------------------------------------------------------------------

    This function processes the provided audio file using the 'pyannote/speaker-diarization' speech diarization
    model, and returns a pandas dataframe containing the speaker diarization information.

    Parameters:
    ...........
    file_path : str
        Path to the input audio file.
    hf_token : str
        Access token for HuggingFace to access pre-trained models.

    Returns:
    ...........
    diart_df : pandas.DataFrame
        A pandas dataframe containing the speaker diarization information.

    ------------------------------------------------------------------------------------------------------
    """
    pipeline = Pipeline.from_pretrained("pyannote/speaker-diarization-3.0", use_auth_token=hf_token)
    diart = pipeline(file_path, num_speakers=2)
    
    diart_df = sutil.get_diart_interval(diart)
    diart_df = diart_df.sort_values(by=['start_time', 'end_time']).reset_index(drop=True)
    return diart_df

def read_kwargs(kwargs):
    """
    ------------------------------------------------------------------------------------------------------

    Reads keyword arguments and returns a dictionary containing input parameters.

    Parameters:
    ...........
    kwargs : dict
        Keyword arguments to be processed.

    Returns:
    ...........
    input_param: dict
        A dictionary containing input parameters with their corresponding values.

    ------------------------------------------------------------------------------------------------------
    """
    input_param = {}
    input_param['hf_token'] = kwargs.get('hf_token', '')
    
    input_param['transcript_json'] = kwargs.get('transcript_json', json.dumps({}))
    input_param['context'] = kwargs.get('context', '')
    input_param['context_model'] = kwargs.get('context_model', 'sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2')

    input_param['volume_normalization'] = kwargs.get('volume_normalization', None)

    return input_param

def get_pyannote(input_param, filepath):
    """
    ------------------------------------------------------------------------------------------------------

    Retrieves speaker identification information using the local Diarization model.

    Parameters:
    ...........
    input_param : dict
        A dictionary containing input parameters
    filepath : str
        The file path.

    Returns:
    ...........
    speaker_df :pandas.DataFrame
        The speaker identification dataframe.
    speaker_count :int
        The number of identified speakers.

    ------------------------------------------------------------------------------------------------------
    """
    
    diart_df = run_pyannote(filepath, input_param['hf_token'])
    transcribe_df = pd.DataFrame(input_param['transcript_json'])

    speaker_df, speaker_count = sutil.get_speaker_identification(diart_df, transcribe_df)
    return speaker_df, speaker_count

def speaker_separation_nolabels(filepath, **kwargs):
    """
    ------------------------------------------------------------------------------------------------------

    This function performs speaker separation using speech diarization techniques on the provided audio file.

    Parameters:
    ...........
    filepath : str
        Path to the input audio file.
    transcript_json : json
        Speech transcription json response.
    hf_token : str
        Access token for HuggingFace to access pre-trained models.
    context : str, optional
        scale to use for slicing the separated audio files, if any.
    context_model : str, optional
        model to use for speaker identification, if context is provided.
    volume_normalization : int
        The volume normalization level. Default is None.

    Returns:
    ...........
    signal_label : pandas.DataFrame
        A pandas dataframe containing the speaker diarization information.

    ------------------------------------------------------------------------------------------------------
    """
    signal_label = {}
    input_param = read_kwargs(kwargs)

    measures = get_config()

    try:
        if not os.path.exists(filepath) or 'transcript_json' not in kwargs:
            return signal_label
        
        if input_param['context_model'] not in measures['embedding_models']:
            raise ValueError(f"Model {input_param['context_model']} not found in the config file")

        speaker_df, speaker_count = get_pyannote(input_param, filepath)
        audio_signal = AudioSegment.from_file(file = filepath, format = "wav")

        if len(speaker_df)>0 and speaker_count>1:
            combined_df = sutil.combine_turns(speaker_df)
            signal_label = sutil.generate_audio_signal(combined_df, audio_signal, input_param['context'], input_param['context_model'], measures)

            if input_param['volume_normalization']:
                if type(input_param['volume_normalization']) != int or input_param['volume_normalization'] < -60 or input_param['volume_normalization'] > 0:
                    logger.info('Volume normalization value should be an integer between -60 and 0')
                    return signal_label
                signal_label = sutil.adjust_volume(filepath, signal_label, input_param['volume_normalization'])

    except Exception as e:
        logger.info(f'Error in diard processing: {e} & File: {filepath}')

    return signal_label



# author:    Vijay Yadav
# website:   http://www.bklynhlth.com

# import the required packages
import os
import tempfile
import pandas as pd
import numpy as np
import logging
import shutil

from sentence_transformers import SentenceTransformer
from pydub import AudioSegment
from transformers import AutoTokenizer, AutoModel
import torch
import torch.nn.functional as F
from ..commons import to_audio, from_audio, volume_normalization

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

def average_pool(last_hidden_states, attention_mask):
    """
    ------------------------------------------------------------------------------------------------------
    
    This function takes in the last hidden states and attention mask of a transformer model, and computes
    the average pooling of the last hidden states based on the attention mask.
    
    Parameters:
    ...........
    last_hidden_states : torch.Tensor
        The last hidden states of the transformer model.
    attention_mask : torch.Tensor
        The attention mask of the transformer model.
    
    Returns:
    ...........
    last_hidden : torch.Tensor
        The average pooled last hidden states of the transformer model.
    
    ------------------------------------------------------------------------------------------------------
    """
    last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)
    return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]

def get_embeddings(model, tokenizer, sentences):
    if tokenizer is None:
        embeddings = model.encode(sentences, normalize_embeddings=True)
        embeddings = np.array(embeddings)
    else:
        sentences[0] = f'Instruct: Retrieve semantically similar text.\nQuery: {sentences[0]}'

        embeddings = []
        batch_size = 8 
        for i in range(0, len(sentences), batch_size):
            sen_list2 = sentences[i:i+batch_size]
            batch_dict = tokenizer(sen_list2, max_length=512, padding=True, truncation=True, return_tensors='pt')

            with torch.no_grad():
                outputs = model(**batch_dict)
            embeddings2 = average_pool(outputs.last_hidden_state, batch_dict['attention_mask'])

            embeddings2 = F.normalize(embeddings2, p=2, dim=1)
            if len(embeddings) == 0:
                embeddings = embeddings2
            else:
                embeddings = torch.cat((embeddings, embeddings2), dim=0)

            del outputs, embeddings2
            torch.cuda.empty_cache()

        embeddings = embeddings.cpu().detach().numpy()

    return embeddings

def match_transcript(sigma_string, speech, model, tokenizer):
    """
    ------------------------------------------------------------------------------------------------------

    The function uses a pre-trained sentence transformer model to compute the similarity between
    the speech and a list of pre-defined script sentences. It
    returns the average similarity score of the top 5 matches.

    Parameters:
    ...........
    sigma_string : str
        a string of sigma script
    speech : str
        a string containing the speech to be matched with the script sentences
    model : SentenceTransformer or AutoModel
        a pre-trained sentence transformer model
    tokenizer : AutoTokenizer
        a pre-trained tokenizer for the model

    Returns:
    ...........
    match_score : float
        a float value representing the average similarity score of the top 5 matches

    ------------------------------------------------------------------------------------------------------
    """
    sigma_script = sigma_string.split(',')
    sen_list = [speech] + sigma_script

    sentence_embeddings = get_embeddings(model, tokenizer, sen_list)
    probs = sentence_embeddings[:1] @ sentence_embeddings[1:].T

    top_5 = np.argsort(probs[0])[-5:]
    match_score = np.mean(probs[0][top_5])
    return match_score

def prepare_diart_interval(start_time, end_time, speaker_list):
    """
    ------------------------------------------------------------------------------------------------------

    The function checks for overlapping intervals and merges them if necessary.

    Parameters:
    ...........
    start_time : list
        a list of start times as floats
    end_time : list
        a list of end times as floats
    speaker_list : list
        a list of speaker names as strings

    Returns:
    ...........
    df : pandas dataframe
        a dataframe containing the start time, end time, interval, and speaker columns

    ------------------------------------------------------------------------------------------------------
    """
    df = pd.DataFrame(start_time, columns=['start_time'])

    df['end_time'] = end_time
    df['interval'] =df['end_time'] - df['start_time']
    
    df['speaker'] = speaker_list
    return df

def get_diart_interval(diarization):
    """
    ------------------------------------------------------------------------------------------------------

    This function takes in a pyannote.core.SegmentManification object representing the diarization results
    and returns a pandas dataframe with columns for start time, end time, interval, and speaker.

    Parameters:
    ...........
    diarization : pyannote.core.SegmentManification
        a diarization object representing the diarization results

    Returns:
    ...........
    df : pandas dataframe
        a dataframe containing the start time, end time, interval, and speaker columns

    ------------------------------------------------------------------------------------------------------
    """
    start_time = []
    end_time = []
    speaker_list = []

    for turn, _, speaker in diarization.itertracks(yield_label=True):
        try:

            speaker_id = str(speaker).split('_')[1]
            speaker_id = int(speaker_id)
            start_time.append(turn.start)

            end_time.append(turn.end)
            speaker_list.append('speaker'+ str(speaker_id))

        except Exception as e:
            logger.info(f'Error in pyannote filtering: {e}')

    df = prepare_diart_interval(start_time, end_time, speaker_list)
    return df

def get_embedding_model(context_model):
    """
    ------------------------------------------------------------------------------------------------------

    This function takes in a string containing the name of the embedding model to be used and returns
    the corresponding pre-trained model and tokenizer.

    Parameters:
    ----------
    context_model : str
        A string containing the name of the embedding model to be used.

    Returns:
    -------
    model : SentenceTransformer or AutoModel
        A pre-trained sentence transformer model.
    tokenizer : AutoTokenizer
        A pre-trained tokenizer for the model.

    ------------------------------------------------------------------------------------------------------
    """
    if context_model == 'sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2':
        model = SentenceTransformer(context_model)
        tokenizer = None
    else:
        tokenizer = AutoTokenizer.from_pretrained(context_model)
        model = AutoModel.from_pretrained(context_model)

    return model, tokenizer

def get_patient_rater_label(df, measures, scale, signal, context_model):
    """
    ------------------------------------------------------------------------------------------------------

    This function takes in a pandas dataframe 'df' containing diarization results and returns a dictionary
    signal_label with the labels assigned to the speakers based on the comparison of their scores.

    Parameters:
    ----------
    df : pandas DataFrame
        A dataframe containing the diarization results
    measures : dict
        A dictionary with config values.
    scale : str
        A clinical scale.
    signal : list
        A list of audio signals.
    context_model : str
        A string containing the name of the embedding model to be used.

    Returns:
    -------
    signal_label : dict
        A dictionary with the assigned labels for the speakers.

    ------------------------------------------------------------------------------------------------------
    """
    signal_label = {}
    spk1_txt = ' '.join(df[df['speaker_label'] == 'speaker0'].reset_index(drop=True)['content'])
    spk2_txt = ' '.join(df[df['speaker_label'] == 'speaker1'].reset_index(drop=True)['content'])

    if scale.lower() not in measures['scale'].split(","):
        return signal

    elif spk1_txt == '' and spk2_txt == '': #Check empty text
        return signal
    
    score_string = scale.lower()+'_string'

    model, tokenizer = get_embedding_model(context_model)

    spk1_score = match_transcript(measures[score_string], spk1_txt, model, tokenizer)
    spk2_score = match_transcript(measures[score_string], spk2_txt, model, tokenizer)

    signal_label = {'clinician': signal['speaker1'], 'participant':signal['speaker0']}

    if spk1_score > spk2_score:
        signal_label = {'clinician': signal['speaker0'], 'participant':signal['speaker1']}
    return signal_label

def get_segment_signal(audio_signal, df):
    """
    ------------------------------------------------------------------------------------------------------

    Extracts speaker-specific segments from an audio signal based on a dataframe.

    Parameters:
    ----------
    audio_signal : AudioSignal
        The audio signal.
    df : pandas DataFrame
        The dataframe containing speaker information.

    Returns:
    -------
    signal_dict: dict
        A dictionary with key as labled speaker and value as a List of numpy arrays representing the audio segments
        of that speaker.

    ------------------------------------------------------------------------------------------------------
    """
    signal_dict = {}

    for index, row in df.iterrows():
        start_time = row['start_time']

        end_time = row['end_time']
        speaker_label = row['speaker_label']

        speaker_audio = audio_signal[start_time:end_time]
        speaker_array = np.array(speaker_audio.get_array_of_samples())

        if speaker_label in ['speaker0', 'speaker1', 'clinician', 'participant']:
            signal_dict.setdefault(speaker_label, []).append(speaker_array)

    return signal_dict

def generate_audio_signal(df, audio_signal, scale, context_model, measures):
    """
    ------------------------------------------------------------------------------------------------------

    Generates a labeled audio signal based on the given DataFrame, audio signal, scale, and measures (common function).

    Parameters:
    ----------
    df : pandas DataFrame
        The DataFrame containing the start and end times of segments.
    audio_signal : AudioSignal
        The original audio signal.
    scale : str
        A clinical scale
    context_model : str
        A string containing the name of the embedding model to be used.
    measures : dict
        A config dictionary.

    Returns:
    -------
    signal_label: dict
        A dictionary with the assigned labels for the speakers and audio signals.

    ------------------------------------------------------------------------------------------------------
    """
    df['start_time'] = df['start_time'].astype(float) *1000
    df['end_time'] = df['end_time'].astype(float)*1000

    signal_dict = get_segment_signal(audio_signal, df)
    signal_dict = {key: np.concatenate(value) for key, value in signal_dict.items()}

    signal_label = get_patient_rater_label(df, measures, scale, signal_dict, context_model)
    return signal_label

def get_speaker_identification(df1, df2):
    """
    ------------------------------------------------------------------------------------------------------

    Identifies speakers from two dataframes based on time intervals (common function).

    Parameters:
    ----------
    df1 : pandas DataFrame
        First dataframe containing speaker information.
    df2 : pandas.DataFrame
        Second dataframe containing time interval information.

    Returns:
    -------
    grouped_df : pandas.DataFrame
        A dataframe with aggregated speaker information.
    speaker_count : int
        The number of unique speaker labels identified.

    ------------------------------------------------------------------------------------------------------
    """
    # Merge the dataframes based on overlapping time intervals
    merged_df = pd.merge(df1, df2, how='cross')

    # Filter the merged dataframe based on overlapping intervals
    filtered_df = merged_df[ (merged_df['start'] >= merged_df['start_time']) &
                            (merged_df['end'] <= merged_df['end_time'])]

    # Groupby and aggregate word and conf
    filtered_df = filtered_df.rename(columns={'conf': 'confidence', 'speaker': 'speaker_label', 'word': 'content', 'start': 'start_time', 'end': 'end_time'}).iloc[:, 2:]

    filtered_df = filtered_df[["start_time", "end_time", "confidence", "speaker_label", "content"]]
    speaker_count = filtered_df['speaker_label'].nunique()
    return filtered_df, speaker_count

def transcribe_response_to_dataframe(response):
    """
    ------------------------------------------------------------------------------------------------------

    Transcribes(cloud:AWS) a response from a speech-to-text service into a pandas DataFrame.

    Parameters:
    ----------
    response : dict
        The response object containing the transcribed data.

    Returns:
    -------
    df : pandas DataFrame
        The transcribed data in a DataFrame.
    speakers: int
        The number of speakers detected in the transcription.

    ------------------------------------------------------------------------------------------------------
    """
    speakers = 0
    df = pd.DataFrame()

    if 'results' in response:
        if 'speaker_labels' in response['results']:

            if 'speakers' in response['results']['speaker_labels']:
                speakers = response['results']['speaker_labels']["speakers"]

            if 'items' in response['results']:
                items = response['results']["items"]
                df = pd.DataFrame(items)

                df["confidence"] = df["alternatives"].apply(lambda x: x[0]["confidence"])
                df["content"] = df["alternatives"].apply(lambda x: x[0]["content"])
                df["confidence"] = df["confidence"].astype(float)

                df = df[df["confidence"] > 0].reset_index(drop=True)
                df = df[["start_time", "end_time", "confidence", "speaker_label", "content"]]
                
    return df, speakers

def adjust_volume(audio_path, signal_label, volume_level):
    """
    ------------------------------------------------------------------------------------------------------

    This function adjusts the volume level of the audio signal.

    Parameters:
    ...........
    audio_path : str
        Path to the input audio file.
    signal_label : pandas.DataFrame
        A pandas dataframe containing the speaker diarization information.
    volume_level : int
        The volume normalization level.

    Returns:
    ...........
    signal_label : pandas.DataFrame
        A pandas dataframe containing the speaker diarization information.

    ------------------------------------------------------------------------------------------------------
    """
    temp_dir = tempfile.mkdtemp()
    to_audio(audio_path, signal_label, temp_dir)
    for file in os.listdir(temp_dir):
        try:
            # standardize volume level
            audio_signal = AudioSegment.from_file(file = os.path.join(temp_dir, file), format = "wav")
            audio_signal = volume_normalization(audio_signal, volume_level)
            audio_signal.export(os.path.join(temp_dir, file), format="wav")
        except Exception as e:
            logger.info(f'Error in adjusting volume for file: {file}, error: {e}')

    signal_label = from_audio(temp_dir)

    # clear the temp directory
    shutil.rmtree(temp_dir)

    return signal_label

def combine_turns(df):
    """
    ------------------------------------------------------------------------------------------------------

    Combines consecutive words of the same speaker into a single turn.

    Parameters:
    ...........
    df : pandas.DataFrame
        The dataframe containing the speaker diarization information.

    Returns:
    ...........
    combined_df : pandas.DataFrame
        The dataframe with combined turns.

    ------------------------------------------------------------------------------------------------------
    """
    change_mask = df['speaker_label'] != df['speaker_label'].shift()
    
    groups = np.cumsum(change_mask)
    
    combined_df = df.groupby(groups).agg({
        'start_time': 'first',
        'end_time': 'last',
        'confidence': 'mean',
        'speaker_label': 'first',
        'content': ' '.join
    }).reset_index(drop=True)
    
    return combined_df



# author:    Georgios Efstathiadis
# website:   http://www.bklynhlth.com

# import the required packages
import os
import json
import logging

from .util import diarization_utils as dutil
from .util import huggingface_utils as hutil

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


def get_config():
    """
    ------------------------------------------------------------------------------------------------------

    Load the configuration settings for the speech transcription.

    Parameters:
    ...........
    None

    Returns:
    ...........
    measures : dict
        A dictionary containing the configuration settings.

    ------------------------------------------------------------------------------------------------------
    """
    #Loading json config
    dir_name = os.path.dirname(os.path.abspath(__file__))
    measure_path = os.path.abspath(os.path.join(dir_name, 'config/speech.json'))

    file = open(measure_path)
    measures = json.load(file)
    return measures


def read_kwargs(kwargs):
    """
    ------------------------------------------------------------------------------------------------------

    Reads keyword arguments and returns a dictionary containing input parameters.

    Parameters:
    ...........
    kwargs : dict
        Keyword arguments to be processed.

    Returns:
    ...........
    input_param: dict A dictionary containing input parameters with their corresponding values.

    ------------------------------------------------------------------------------------------------------
    """
    input_param = {}
    input_param['huggingface_token'] = kwargs.get('huggingface_token', '')

    return input_param


def is_amazon_transcribe(transcript_json):
    """
    ------------------------------------------------------------------------------------------------------
    This function checks if the json response object is from Amazon Transcribe.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.

    Returns:
    ...........
    bool: True if the json response object
     is from Amazon Transcribe, False otherwise.

    ------------------------------------------------------------------------------------------------------
    """
    return "jobName" in transcript_json and "results" in transcript_json


def is_whisper_transcribe(transcript_json):
    """
    ------------------------------------------------------------------------------------------------------

    This function checks if the json response object is from Whisper Transcribe.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.

    Returns:
    ...........
    bool: True if the json response object
     is from Whisper Transcribe, False otherwise.

    ------------------------------------------------------------------------------------------------------
    """
    if "segments" in transcript_json:
        if len(transcript_json["segments"]) > 0:

            if "words" in transcript_json["segments"][0]:
                return True
    return False


def diarization_correction(transcript_json, model_name, **kwargs):
    """
    ------------------------------------------------------------------------------------------------------
    This function corrects the speaker diarization of a transcript
     using a fine-tuned speaker diarization LLM.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.
    model_name: str
        The name of the model used for transcription in HuggingFace.
    kwargs: dict
        Additional arguments for the AWS API call.

    Returns:
    ...........
    dict: JSON response object with corrected speaker diarization.

    ------------------------------------------------------------------------------------------------------
    """
    transcript_json_corrected = transcript_json.copy()
    input_param = read_kwargs(kwargs)
    measures = get_config()

    try:

        if bool(transcript_json):

            if is_whisper_transcribe(transcript_json):
                asr = "whisperx"
            elif is_amazon_transcribe(transcript_json):
                asr = "aws"
            else:
                raise Exception("Transcript source not supported")

            prompts, translate_json = dutil.extract_prompts(transcript_json, asr)
            results = hutil.call_diarization_hf(prompts, model_name, input_param)
            transcript_json_corrected = dutil.correct_transcription(
                transcript_json, prompts, results, translate_json, asr
            )

    except Exception as e:
        logger.info(f"Error in Speaker Diarization Correction {e}")

    return transcript_json_corrected



# website:   http://www.brooklyn.health

# import the required packages
import os
import wave
import json
import logging
import json

from vosk import Model, KaldiRecognizer
from pydub import AudioSegment
from .util import util as ut

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

def get_config():
    """
    ------------------------------------------------------------------------------------------------------

    Load the configuration settings for the speech transcription.

    Parameters:
    ...........
    None

    Returns:
    ...........
    measures : dict
        A dictionary containing the configuration settings.

    ------------------------------------------------------------------------------------------------------
    """
    #Loading json config
    dir_name = os.path.dirname(os.path.abspath(__file__))
    measure_path = os.path.abspath(os.path.join(dir_name, 'config/speech.json'))

    file = open(measure_path)
    measures = json.load(file)
    return measures

def filter_audio(filepath, t_interval):
    """
    ------------------------------------------------------------------------------------------------------

    Filter an audio file to extract a segment based on the specified time interval.

    Parameters:
    ............
    filepath : str
        The path to the audio file to be filtered.
    t_interval : list
        A list of tuples representing the start and end times (in seconds) of the segment to extract.

    Returns:
    ............
    sound : AudioSegment
        The filtered audio segment.

    ------------------------------------------------------------------------------------------------------
    """
    sound = AudioSegment.from_wav(filepath)

    if len(t_interval)==2:
        sound = sound[int(t_interval[0])*1000 : int(t_interval[1])*1000]

    elif len(t_interval)==1:
        sound = sound[int(t_interval[0])*1000:]

    sound = sound.set_channels(1)
    return sound

def filter_speech(measures, results):
    """
    ------------------------------------------------------------------------------------------------------

    Filter the speech transcription results to extract the transcript.

    Parameters:
    ...........
    measures : dict
        A dictionary containing the configuration settings for the speech transcription.
    results : list of dict
        The raw transcription results returned by the transcription service.

    Returns:
    ...........
    result_key : list
        A list containing the framewise transcription of the audio file.
    transcript : str
        The transcript of the audio file.

    ------------------------------------------------------------------------------------------------------
    """
    result_key = []
    text_key = []
    transcript_dict = {}

    for res in results:
        dict_keys = res.keys()

        if 'result' in dict_keys and 'text' in dict_keys:
            result_key.extend(res['result'])
            text_key.append(res['text'])

    transcript_dict['result'] = result_key
    transcript_dict['text'] = ' '.join(text_key)
    return result_key, ' '.join(text_key)

def get_vosk(audio_path, lang):
    """
    ------------------------------------------------------------------------------------------------------

    Recognize speech using the Vosk model.

    Parameters:
    ............
    audio_path : str
        The path to the audio file to be transcribed.
    lang : str
        The language of the audio file (e.g. 'en-us', 'es', 'fr').

    Returns:
    ............
    results : list of dict
        The raw transcription results returned by the Vosk model.

    ------------------------------------------------------------------------------------------------------
    """
    model = Model(lang=lang)
    wf = wave.open(audio_path, "rb")

    recog = KaldiRecognizer(model, wf.getframerate())
    recog.SetWords(True)

    results = []
    while True:

        data = wf.readframes(4000) #Future work
        if len(data) == 0:
            break

        if recog.AcceptWaveform(data):
            partial_result = json.loads(recog.Result())
            results.append(partial_result)

    partial_result = json.loads(recog.FinalResult())
    results.append(partial_result)
    return results

def stereo_to_mono(filepath, t_interval):
    """
    ------------------------------------------------------------------------------------------------------

    Convert a stereo audio file to a mono audio file.

    Parameters:
    ............
    filepath : str
        The path to the stereo audio file to be converted.
    t_interval : list
        A list of tuples representing the start and end times (in seconds) of segments of the audio file to be transcribed.

    Returns:
    ............
    mono_filepath : str
        The path to the mono audio file.

    ------------------------------------------------------------------------------------------------------
    """
    sound = filter_audio(filepath, t_interval)

    filename, _ = os.path.splitext(os.path.basename(filepath))
    dir_name = os.path.join(os.path.dirname(filepath), 'temp_mono_' + filename)

    ut.make_dir(dir_name)
    mono_filepath = os.path.join(dir_name, filename + '.wav')
    sound.export(mono_filepath, format="wav")
    return mono_filepath

def run_vosk(filepath, language, transcribe_interval = []):
    """
    ------------------------------------------------------------------------------------------------------

    Transcribe speech in an audio file using the Vosk model.

    Parameters:
    ............
    filepath : str
        The path to the audio file to be transcribed.
    language : str, optional
        The language of the audio file (e.g. 'en-us', 'es', 'fr'). Default is 'en-us'.
    transcribe_interval : list, optional
        A list of tuples representing the start and end times (in seconds) of segments of the audio file to be transcribed.
        Default is an empty list.

    Returns:
    ............
    json_response : str
        The JSON response from the Vosk transcription service.
    transcript : str
        The transcript of the audio file.

    ------------------------------------------------------------------------------------------------------
    """
    json_response = json.dumps({})
    transcript = mono_filepath = ''

    try:
        if os.path.exists(filepath):

            measures = get_config()
            mono_filepath = stereo_to_mono(filepath, transcribe_interval)
            results = get_vosk(mono_filepath, language)

            ut.remove_dir(os.path.dirname(mono_filepath)) #Clean temp directory
            json_response, transcript = filter_speech(measures, results)

        else:
            logger.info(f'Audio file not available. File: {filepath}')

    except Exception as e:
        ut.remove_dir(os.path.dirname(mono_filepath))#Clean temp directory
        logger.info(f'Error in speech Transcription: {e} & File: {filepath}')

    finally:
        return json_response, transcript

    

def speech_transcription_vosk(filepath, **kwargs):
    """
    ------------------------------------------------------------------------------------------------------

    Speech transcription function that transcribes an audio file using vosk.

    Parameters:
    ...........
    filepath : str
        The path to the audio file to be transcribed.
    language : str, optional
        The language of the audio file (e.g. 'en-us', 'es', 'fr'). Default is 'en-us'.
    transcribe_interval : list, optional
        A list of tuples representing the start and end times (in seconds) of segments of the audio file to be transcribed.
        Only applicable if model is 'vosk'. Default is an empty list.

    Returns:
    ...........
    json_response : JSON Object
        A transcription response object in JSON format
    transcript : str
        The transcript of the recording.

    ------------------------------------------------------------------------------------------------------
    """

    measures = get_config()
    language = kwargs.get('language', 'en-us')
    transcribe_interval = kwargs.get('transcribe_interval', [])
    
    json_response, transcript = run_vosk(filepath, language, transcribe_interval)
    return json_response, transcript



# author:    Vijay Yadav
# website:   http://www.bklynhlth.com

# import the required packages
import os
import json
import logging

from .util import separation_util as sutil
from .commons import whisperx_to_dataframe, transcribe_response_to_dataframe
from pydub import AudioSegment

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

def get_config():
    """
    ------------------------------------------------------------------------------------------------------

    This function loads and returns a dictionary containing the configuration settings for speech separation
    from a JSON file.

    Parameters:
    ...........
    None

    Returns:
    ...........
    measures : dict
        A dictionary containing the configuration settings for speech separation.

    ------------------------------------------------------------------------------------------------------
    """
    #Loading json config
    dir_name = os.path.dirname(os.path.abspath(__file__))
    measure_path = os.path.abspath(os.path.join(dir_name, 'config/speech.json'))

    file = open(measure_path)
    measures = json.load(file)
    return measures

def is_amazon_transcribe(json_conf):
    """
    ------------------------------------------------------------------------------------------------------
    This function checks if the json response object is from Amazon Transcribe.
    Parameters:
    ...........
    json_conf: dict
        JSON response object.
    Returns:
    ...........
    bool: True if the json response object
     is from Amazon Transcribe, False otherwise.
    ------------------------------------------------------------------------------------------------------
    """
    return "jobName" in json_conf and "results" in json_conf

def speaker_separation_labels(filepath, transcript_json, volume_normalization=None):
    """
    ------------------------------------------------------------------------------------------------------

    This function performs speaker separation using cloud based speech diarization techniques.

    Parameters:
    ...........
    filepath : str
        Path to the input audio file.
    transcript_json : json
        Speech transcription json response.
    volume_normalization : int
        The volume normalization level. Default is None.

    Returns:
    ...........
    signal_label : pandas.DataFrame
        A pandas dataframe containing the speaker diarization information.

    ------------------------------------------------------------------------------------------------------
    """
    signal_label = {}
    measures = get_config()

    try:
        if not os.path.exists(filepath):
            return signal_label

        audio_signal = AudioSegment.from_file(file = filepath, format = "wav")
        if not is_amazon_transcribe(transcript_json):
            speaker_df, speaker_count = whisperx_to_dataframe(transcript_json)
        else:
            speaker_df, speaker_count = transcribe_response_to_dataframe(transcript_json)
            
        if len(speaker_df)>0 and speaker_count>1:
            combined_df = sutil.combine_turns(speaker_df)
            signal_label = sutil.generate_audio_signal(combined_df, audio_signal, '', '', measures)

            if volume_normalization:
                if type(volume_normalization) != int or volume_normalization < -60 or volume_normalization > 0:
                    logger.info('Volume normalization value should be an integer between -60 and 0')
                    return signal_label
                signal_label = sutil.adjust_volume(filepath, signal_label, volume_normalization)

    except Exception as e:
        logger.info(f'Error in diard processing: {e} & File: {filepath}')

    return signal_label



# author:    Vijay Yadav
# website:   http://www.bklynhlth.com

import whisperx
import gc 
import torch

import json
import logging

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()

def delete_model(model):
    """
    ------------------------------------------------------------------------------------------------------

    delete model if low on GPU resources
    Parameters:
    ...........
    model : object
        loaded model object
    
    ------------------------------------------------------------------------------------------------------
    """
    gc.collect()
    torch.cuda.empty_cache()
    del model

def get_diarization(audio, align_json, device, input_param):
    """
    ------------------------------------------------------------------------------------------------------

    Assign speaker labels
    Parameters:
    ...........
    audio : object
        audio signal object
    align_json: json
        aligned whisper transcribed output
    device : str
        device type
    input_param : dict
        A dictionary containing input parameters
    
    Returns:
    ...........
    json_response : JSON Object
        A transcription response object in JSON format

    ------------------------------------------------------------------------------------------------------
    """
    # Assign speaker labels
    if input_param['infra_model'][0]:
        diarize_model = whisperx.DiarizationPipeline(use_auth_token=input_param['hf_token'], device=device)
    else:
        diarize_model = input_param['infra_model'][2]

    if input_param['min_speakers'] == None and input_param['max_speakers'] == None:
        diarize_segments = diarize_model(audio)
    
    elif input_param['min_speakers'] == None and input_param['max_speakers'] != None:
        diarize_segments = diarize_model(audio, max_speakers = input_param['max_speakers'])
    
    elif input_param['min_speakers'] != None and input_param['max_speakers'] == None:
        diarize_segments = diarize_model(audio, min_speakers= input_param['min_speakers'])
        
    else:
        diarize_segments = diarize_model(audio, min_speakers=input_param['min_speakers'], max_speakers=input_param['max_speakers'])
        
    json_response = whisperx.assign_word_speakers(diarize_segments, align_json)
    return json_response

def get_transcribe_summary(json_response):
    """
    ------------------------------------------------------------------------------------------------------

    Assign speaker labels
    Parameters:
    ...........
    json_response: json
        whisper transcribed output
    
    Returns:
    ...........
    summary : str
        whisper transcribed test summary

    ------------------------------------------------------------------------------------------------------
    """
    summary = ""
    
    if 'segments' in json_response:
        summary = "".join([item['text'] for item in json_response["segments"] if item.get('text', '')])
    return summary

def transcribe_whisper(filepath, model, device, compute_type, batch_size, infra_model, language):
    """
    ------------------------------------------------------------------------------------------------------
   
    Transcribe with whisper (batched)
    
    Parameters:
    ...........
    filepath : str
        The path to the audio file to be transcribed.
    model: str
        name of the pretrained model
    device: str
        cpu vs gpu
    compute_type: str
        computation format
    batch_size: int
        batch size
    infra_model:list
        whisper model artifacts (this is optional param: to optimize willisInfra) 
    language: str
        language code

        
    ------------------------------------------------------------------------------------------------------
    """
    if infra_model[0]:
        model_whisp = whisperx.load_model(model, device, compute_type=compute_type)
    else:
        model_whisp = infra_model[1] #passing param from willismeansure
    audio = whisperx.load_audio(filepath)

    transcribe_json = model_whisp.transcribe(audio, batch_size=batch_size, language=language)
    return transcribe_json, audio

def get_whisperx_diarization(filepath, input_param):
    """
    ------------------------------------------------------------------------------------------------------

    Transcribe an audio file using Whisperx.

    Parameters:
    ...........
    filepath : str
        The path to the audio file to be transcribed.
    input_param : dict
        A dictionary containing input parameters

    Returns:
    ...........
    json_response : JSON Object
        A transcription response object in JSON format
    transcript : str
        The transcript of the recording.

    ------------------------------------------------------------------------------------------------------
    """
    device = input_param['device_type']
    compute_type = input_param['compute_type']
    
    json_response = json.dumps({})
    transcript = ''
    
    try:
        if not torch.cuda.is_available() and device == 'cuda':
            raise Exception('CUDA is not available')

        transcribe_json, audio = transcribe_whisper(filepath, input_param['model'], device, compute_type, input_param['batch_size'], input_param['infra_model'], input_param['language'])
    
        # Align whisper output
        model_a, metadata = whisperx.load_align_model(language_code=input_param['language'], device=device)
        align_json = whisperx.align(transcribe_json["segments"], model_a, metadata, audio, device, return_char_alignments=False)
    
        if input_param['del_model']:
            delete_model(model_a)
            
        json_response = get_diarization(audio, align_json, device, input_param)    
        transcript = get_transcribe_summary(json_response)
    
    except Exception as e:
        logger.info(f'Error in speech Transcription: {e} & File: {filepath}')
    return json_response, transcript



# author:    Vijay Yadav
# website:   http://www.bklynhlth.com

# import the required packages

import os
import json
import logging

from .commons import get_config
from .util import transcribe_util as tutil
from .willisdiarize import diarization_correction

logging.basicConfig(level=logging.INFO)
logger=logging.getLogger()


def read_kwargs(kwargs):
    """
    ------------------------------------------------------------------------------------------------------

    Reads keyword arguments and returns a dictionary containing input parameters.

    Parameters:
    ...........
    kwargs : dict
        Keyword arguments to be processed.

    Returns:
    ...........
    input_param: dict
        A dictionary containing input parameters with their corresponding values.

    ------------------------------------------------------------------------------------------------------
    """
    input_param = {}
    input_param['model'] = kwargs.get('model', 'tiny')
    input_param['language'] = kwargs.get('language', 'en')
    
    input_param['context'] = kwargs.get('context', '')
    input_param['context_model'] = kwargs.get('context_model', 'sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2')
    input_param['max_speakers'] = kwargs.get('max_speakers', None)
    input_param['min_speakers'] = kwargs.get('min_speakers', None)

    input_param['hf_token'] = kwargs.get('hf_token', '')
    input_param['del_model'] = kwargs.get('del_model', False) #Temp filter
    input_param['infra_model'] = kwargs.get('infra_model', [True, None, None]) #Temp filter
    input_param['compute_type'] = kwargs.get('compute_type', 'int16')
    input_param['device_type'] = kwargs.get('device_type', 'cpu')
    input_param['batch_size'] = kwargs.get('batch_size', 16)

    input_param['willisdiarize'] = kwargs.get('willisdiarize', '')

    return input_param

def run_whisperx(filepath, input_param):
    """
    ------------------------------------------------------------------------------------------------------

    Transcribe audio data using the WhisperX model.

    Parameters:
    ...........
    filepath : str
        The path to the audio file to be transcribed.
    input_param : dict
        A dictionary containing input parameters

    Returns:
    ...........
    json_response : JSON Object
        A transcription response object in JSON format
    transcript : str
        The transcript of the recording.

    ------------------------------------------------------------------------------------------------------
    """
    json_response = json.dumps({})
    transcript = ''
    
    if os.path.exists(filepath)== False or input_param['hf_token'] == '':
        return json_response, transcript
    
    from .util import whisperx_util as wutil #import in-case of model=whisperx
    json_response, transcript = wutil.get_whisperx_diarization(filepath, input_param)
    
    if str(json_response) != '{}':
        json_response = tutil.filter_labels_whisper(json_response)
    return json_response, transcript
    

def speech_transcription_whisper(filepath, **kwargs):
    """
    ------------------------------------------------------------------------------------------------------

    Speech transcription function that transcribes an audio file using whisperx.

    Parameters:
    ...........
    filepath : str
        The path to the audio file to be transcribed.
    model : str, optional
        The transcription model to use ('vosk'). Default is 'vosk'.
    language : str, optional
        The language of the audio file (e.g. 'en-us', 'es', 'fr'). Default is 'en-us'.
    transcribe_interval : list, optional
        A list of tuples representing the start and end times (in seconds) of segments of the audio file to be transcribed.
        Only applicable if model is 'vosk'. Default is an empty list.

    Returns:
    ...........
    json_response : JSON Object
        A transcription response object in JSON format
    transcript : str
        The transcript of the recording.

    ------------------------------------------------------------------------------------------------------
    """
    measures = get_config(os.path.abspath(__file__), 'speech.json')
    input_param = read_kwargs(kwargs)

    if not os.path.exists(filepath):
        logger.error("File path does not exist")
        return {}, ''
    
    json_response, transcript = run_whisperx(filepath, input_param)

    if input_param['language'].lower()[:2] == 'en' and input_param['willisdiarize'] != '':
        json_response = diarization_correction(json_response, input_param['willisdiarize'], huggingface_token=input_param['hf_token'])

    if input_param['context'].lower() in measures['scale'].split(',') and input_param['context_model'] in measures['embedding_models']:
        content_dict = tutil.get_whisperx_content(json_response)
        json_response = tutil.get_whisperx_clinical_labels(input_param['context'], measures, content_dict, json_response, input_param['context_model'])

    return json_response, transcript



# author:    Georgios Efstathiadis
# website:   http://www.bklynhlth.com

# import the required packages
import os
import json
from enum import Enum
from multiprocessing import Pool
import time
import random
import logging
from functools import wraps

import boto3
import numpy as np
from scipy import optimize

# avoid warning about tokenizers parallelism
os.environ["TOKENIZERS_PARALLELISM"] = "false"

PROMPT_FORMAT = """### Instruction
In the speaker diarization transcript below, some words are potentially misplaced. Please correct those words and move them to the right speaker. Directly show the corrected transcript without explaining what changes were made or why you made those changes.:

{{ user_msg_1 }}

### Answer

"""


def exponential_backoff_decorator(max_retries, base_delay):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            retries = 0
            while retries < max_retries:
                try:
                    result_func = func(*args, **kwargs)
                    return result_func
                except Exception as e:
                    retries += 1
                    delay = base_delay * 2**retries + random.uniform(0, 1)
                    logging.error(f"Retrying in {delay:.2f} seconds...")
                    time.sleep(delay)
            raise Exception("Max retries reached, operation failed.")

        return wrapper

    return decorator

### Functions taken from https://github.com/google/speaker-id/tree/master/DiarizationLM
### Slightly modified to fit OpenWillis
def create_diarized_text(word_labels, speaker_labels):
    """
    ------------------------------------------------------------------------------------------------------

    This function creates a diarized text from word and speaker labels.

    Parameters:
    ...........
    word_labels: list
        List of words.
    speaker_labels: list
        List of speaker labels.

    Returns:
    ...........
    str: Diarized text.
     e.g. "<spk:1> hello <spk:2> how are you <spk:1> I am fine"

    ------------------------------------------------------------------------------------------------------
    """
    output = []
    previous_speaker = None

    for word, speaker in zip(word_labels, speaker_labels):
        if speaker != previous_speaker:
            output.append("<spk:" + speaker + ">")

        output.append(word)
        previous_speaker = speaker

    return " ".join(output)


def extract_text_and_spk(diarized_text):
    """
    ------------------------------------------------------------------------------------------------------

    This function extracts text and speaker labels from diarized text.

    Parameters:
    ...........
    diarized_text: str
        Previously diarized string.

    Returns:
    ...........
    str: words concatenated with spaces.
    str: speaker labels concatenated with spaces.

    ------------------------------------------------------------------------------------------------------
    """
    spk = "1"
    previous_spk = "1"
    result_text = []
    result_spk = []

    for word in diarized_text.split():
        if word.startswith("<spk:"):
            if not word.endswith(">"):
                word += ">"
            spk = word[len("<spk:") : -len(">")]
            # Handle undefined behaviors of non-recognizable spk with a placeholder.
            try:
                spk_int = int(spk)
                if not spk or spk_int < 1 or spk_int > 10:
                    raise ValueError("Seeing unexpected word: ", word)
                previous_spk = spk
            except ValueError:
                print("Skipping meaningless speaker token:", word)
                spk = previous_spk
        else:
            result_text.append(word)
            result_spk.append(spk)

    return " ".join(result_text), " ".join(result_spk)


class EditOp(Enum):
    Correct = 0
    Substitution = 1
    Insertion = 2
    Deletion = 3


# Computes the Levenshtein alignment between strings ref and hyp, where the
# tokens in each string are separated by delimiter.
# Outputs a tuple : (edit_distance, alignment) where
# alignment is a list of pairs (ref_pos, hyp_pos) where ref_pos is a position
# in ref and hyp_pos is a position in hyp.
# As an example, for strings 'a b' and 'a c', the output would look like:
# (1, [(0,0), (1,1)]
# Note that insertions are represented as (-1, j) and deletions as (i, -1).
def levenshtein_with_edits(ref, hyp, delimiter=" ", print_debug_info=False):
    align = []
    s1 = ref.split(delimiter)
    s2 = hyp.split(delimiter)
    n1 = len(s1)
    n2 = len(s2)
    costs = np.zeros((n1 + 1, n2 + 1), dtype=np.int16)
    backptr = np.zeros((n1 + 1, n2 + 1), dtype=EditOp)

    for i in range(n1 + 1):  # ref
        costs[i][0] = i  # deletions

    for j in range(n2):  # hyp
        costs[0][j + 1] = j + 1  # insertions
        for i in range(n1):  # ref
            # (i,j) <- (i,j-1)
            ins = costs[i + 1][j] + 1
            # (i,j) <- (i-1,j)
            del_ = costs[i][j + 1] + 1
            # (i,j) <- (i-1,j-1)
            sub = costs[i][j] + (s1[i] != s2[j])
            costs[i + 1][j + 1] = min(ins, del_, sub)
            if costs[i + 1][j + 1] == ins:
                backptr[i + 1][j + 1] = EditOp.Insertion
            elif costs[i + 1][j + 1] == del_:
                backptr[i + 1][j + 1] = EditOp.Deletion
            elif s1[i] == s2[j]:
                backptr[i + 1][j + 1] = EditOp.Correct
            else:
                backptr[i + 1][j + 1] = EditOp.Substitution

    if print_debug_info:
        print("Mincost: ", costs[n1][n2])
    i = n1
    j = n2
    # Emits pairs (n1_pos, n2_pos) where n1_pos is a position in n1 and n2_pos
    # is a position in n2.
    while i > 0 or j > 0:
        if print_debug_info:
            print("i: ", i, " j: ", j)
        ed_op = EditOp.Correct
        if i >= 0 and j >= 0:
            ed_op = backptr[i][j]
        if i >= 0 and j < 0:
            ed_op = EditOp.Deletion
        if i < 0 and j >= 0:
            ed_op = EditOp.Insertion
        if i < 0 and j < 0:
            raise RuntimeError("Invalid alignment")
        if ed_op == EditOp.Insertion:
            align.append((-1, j - 1))
            j -= 1
        elif ed_op == EditOp.Deletion:
            align.append((i - 1, -1))
            i -= 1
        else:
            align.append((i - 1, j - 1))
            i -= 1
            j -= 1

    align.reverse()
    return costs[n1][n2], align


def normalize_text(text):
    """Normalize text."""
    # Convert to lower case.
    text_lower = text.lower()
    # Remove punctuation.
    text_de_punt = text_lower.replace(",", "").replace(".", "").replace("_", "").strip()
    if len(text_lower.split()) == len(text_de_punt.split()):
        return text_de_punt
    else:
        # If ater removing punctuation, we dropped words, then we keep punctuation.
        return text_lower


def get_aligned_hyp_speakers(
    hyp_text, ref_text, ref_spk,
    print_debug_info=False,
):
    """Align ref_text to hyp_text, then apply the alignment to ref_spk."""
    # Counters for insertions and deletions in hyp and ref text alignment.
    num_insertions, num_deletions = 0, 0

    # Get the alignment.
    _, align = levenshtein_with_edits(
        normalize_text(ref_text), normalize_text(hyp_text)
    )

    ref_spk_list = ref_spk.split()
    hyp_spk_align = []

    # Apply the alignment on ref speakers.
    for i, j in align:
        if i == -1:
            # hyp has insertion
            hyp_spk_align.append("-1")
            num_insertions += 1
        elif j == -1:
            # hyp has deletion
            num_deletions += 1
            continue
        else:
            hyp_spk_align.append(ref_spk_list[i])
    hyp_spk_align = " ".join(hyp_spk_align)

    if print_debug_info:
        print("Number of insertions: ", num_insertions)
        print("Number of deletions: ", num_deletions)
        # This is not the traditional denominator of WER. Instead, this is
        # len(hyp) + len(ref) - len(SUB).
        print("Length of align pairs: ", len(align))
    return hyp_spk_align


def get_oracle_speakers(hyp_spk, hyp_spk_align):
    """Get the oracle speakers for hypothesis."""
    hyp_spk_list = [int(x) for x in hyp_spk.split()]
    hyp_spk_align_list = [int(x) for x in hyp_spk_align.split()]

    # Build cost matrix.
    max_spk = max(max(hyp_spk_list), max(hyp_spk_align_list))
    cost_matrix = np.zeros((max_spk, max_spk))
    for aligned, original in zip(hyp_spk_align_list, hyp_spk_list):
        cost_matrix[aligned - 1, original - 1] += 1

    # Solve alignment.
    row_index, col_index = optimize.linear_sum_assignment(cost_matrix, maximize=True)

    # Build oracle.
    hyp_spk_oracle = hyp_spk_list.copy()
    for i in range(len(hyp_spk_list)):
        if hyp_spk_align_list[i] == -1:
            # There are some missing words. In such cases, we just use the original
            # speaker for these words if possible.
            if hyp_spk_list[i] == -1:
                # If we don't have original speaker for missing words, just use the
                # previous speaker if possible.
                # This is useful for the update_hyp_text_in_utt_dict() function.
                if i == 0:
                    hyp_spk_oracle[i] = 1
                else:
                    hyp_spk_oracle[i] = hyp_spk_oracle[i - 1]
            continue
        assert row_index[hyp_spk_align_list[i] - 1] == hyp_spk_align_list[i] - 1
        hyp_spk_oracle[i] = col_index[hyp_spk_align_list[i] - 1] + 1

    return hyp_spk_oracle


# Transcript-Preserving Speaker Transfer (TPST)
def transcript_preserving_speaker_transfer(src_text, src_spk, tgt_text, tgt_spk):
    """Apply source speakers to target."""

    if len(tgt_text.split()) != len(tgt_spk.split()):
        raise ValueError("tgt_text and tgt_spk must have the same length")

    if len(src_text.split()) != len(src_spk.split()):
        raise ValueError("src_text and src_spk must have the same length")

    tgt_spk_align = get_aligned_hyp_speakers(
        hyp_text=tgt_text,
        ref_text=src_text,
        ref_spk=src_spk,
    )
    oracle_speakers = get_oracle_speakers(hyp_spk=tgt_spk, hyp_spk_align=tgt_spk_align)

    return " ".join([str(x) for x in oracle_speakers])
###


def extract_transcription_aws(transcript_json):
    """
    ------------------------------------------------------------------------------------------------------

    This function extracts the speaker labels and words from an AWS transcription JSON.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.

    Returns:
    ...........
    labels_aws: list
        List of speaker labels.
    words_aws: list
        List of transcribed words.
    translate_json: dict
        Dictionary to translate '1' and '2' to the original speaker labels.

    ------------------------------------------------------------------------------------------------------
    """
    labels_aws = []
    words_aws = []

    for item in transcript_json["results"]["items"]:
        if item["type"] == "pronunciation":
            labels_aws.append(item["speaker_label"])
            words_aws.append(item["alternatives"][0]["content"].lower())

    # get unique speaker labels
    labels_aws_set = list(set(labels_aws))
    if len(list(set(labels_aws))) != 2:
        raise ValueError("Only two speakers supported")

    translate_json = {labels_aws_set[0]: "1", labels_aws_set[1]: "2"}
    # transform speaker labels to '1' and '2'
    for i, label in enumerate(labels_aws):
        labels_aws[i] = translate_json[label]

    # opposite to recreate original speaker labels
    translate_json = {"1": labels_aws_set[0], "2": labels_aws_set[1]}
    return labels_aws, words_aws, translate_json


def extract_transcription_whisperx(transcript_json):
    """
    ------------------------------------------------------------------------------------------------------

    This function extracts the speaker labels and words from a WhisperX transcribed JSON.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.

    Returns:
    ...........
    labels_whisperx: list
        List of speaker labels.
    words_whisperx: list
        List of transcribed words.
    translate_json: dict
        Dictionary to translate '1' and '2' to the original speaker labels.

    ------------------------------------------------------------------------------------------------------
    """
    labels_whisperx = []
    words_whisperx = []

    for item in transcript_json["segments"]:
        labels_whisperx += [item["speaker"] for _ in range(len(item["words"]))]
        words_whisperx += [x["word"].lower() for x in item["words"]]

    # get unique speaker labels
    labels_whisperx_set = list(set(labels_whisperx))
    if len(list(set(labels_whisperx))) != 2:
        raise ValueError("Only two speakers supported")

    translate_json = {labels_whisperx_set[0]: "1", labels_whisperx_set[1]: "2"}
    # transform speaker labels to '1' and '2'
    for i, label in enumerate(labels_whisperx):
        labels_whisperx[i] = translate_json[label]

    # opposite of translate_json
    translate_json = {"1": labels_whisperx_set[0], "2": labels_whisperx_set[1]}
    return labels_whisperx, words_whisperx, translate_json


def modify_transcription_aws(transcript_json, corrected_labels):
    """
    ------------------------------------------------------------------------------------------------------

    This function modifies the speaker labels in an AWS transcription JSON
        using the corrected labels.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.
    corrected_labels: list
        List of corrected speaker labels.

    Returns:
    ...........
    dict: JSON response object with corrected speaker labels.

    ------------------------------------------------------------------------------------------------------
    """

    for item in transcript_json["results"]["items"]:
        if item["type"] == "pronunciation":
            new_label = corrected_labels.pop(0)
            item["speaker_label"] = new_label
        else:
            item["speaker_label"] = new_label

    return transcript_json


def modify_transcription_whisperx(transcript_json, corrected_labels):
    """
    ------------------------------------------------------------------------------------------------------

    This function modifies the speaker labels in an WhisperX transcription JSON
        using the corrected labels.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.
    corrected_labels: list
        List of corrected speaker labels.

    Returns:
    ...........
    dict: JSON response object with corrected speaker labels.

    ------------------------------------------------------------------------------------------------------
    """

    res = []
    words = []
    starts = []
    ends = []

    for item in transcript_json["segments"]:
        for x in item["words"]:
            words.append(x["word"])
            starts.append(x["start"])
            ends.append(x["end"])

    temp = {"words": []}

    for i in range(len(words)):
        if i > 0 and corrected_labels[i] != corrected_labels[i - 1]:

            temp["start"] = temp["words"][0]["start"]
            temp["end"] = temp["words"][-1]["end"]
            temp["speaker"] = temp["words"][0]["speaker"]
            temp["text"] = " ".join([x["word"] for x in temp["words"]])

            res.append(temp)

            temp = {"words": []}

            if i == len(words) - 1:
                temp["words"].append(
                    {
                        "word": words[i],
                        "start": starts[i],
                        "end": ends[i],
                        "speaker": corrected_labels[i],
                    }
                )
                temp["start"] = temp["words"][0]["start"]
                temp["end"] = temp["words"][-1]["end"]
                temp["speaker"] = temp["words"][0]["speaker"]
                temp["text"] = " ".join([x["word"] for x in temp["words"]])

                res.append(temp)

                temp = {"words": []}

        elif i == len(words) - 1 and corrected_labels[i] == corrected_labels[i - 1]:

            temp["words"].append(
                {
                    "word": words[i],
                    "start": starts[i],
                    "end": ends[i],
                    "speaker": corrected_labels[i],
                }
            )
            temp["start"] = temp["words"][0]["start"]
            temp["end"] = temp["words"][-1]["end"]
            temp["speaker"] = temp["words"][0]["speaker"]
            temp["text"] = " ".join([x["word"] for x in temp["words"]])

            res.append(temp)

        temp["words"].append(
            {
                "word": words[i],
                "start": starts[i],
                "end": ends[i],
                "speaker": corrected_labels[i],
            }
        )

    return {"segments": res}


def split_transcription(words, speakers, character_limit=2250):
    """
    ------------------------------------------------------------------------------------------------------

    This function splits a transcription into chunks of predefined character limit.
    
    Parameters:
    ...........
    words:
        List of transcribed words.
    speakers:
        List of speaker labels.
    character_limit:
        Maximum number of characters in each chunk.

    Returns:
    ...........
    prompts:
        List of diarized text chunks.
    """

    prompts = []
    current_chunk_words = []
    current_chunk_speakers = []
    current_chunk_length = 0

    # split the transcription into chunks based on character limit
    for word, speaker in zip(words, speakers):
        if current_chunk_length + len(word) + 1 > character_limit:  # +1 for space or punctuation
            prompts.append(create_diarized_text(current_chunk_words, current_chunk_speakers))
            current_chunk_words = []
            current_chunk_speakers = []
            current_chunk_length = 0

        current_chunk_words.append(word)
        current_chunk_speakers.append(speaker)
        current_chunk_length += len(word) + 1  # +1 for space or punctuation

    if current_chunk_words:  # Add the last chunk if it exists
        prompts.append(create_diarized_text(current_chunk_words, current_chunk_speakers))

    return prompts


def apply_formatting(chunk):
    """
    ------------------------------------------------------------------------------------------------------

    This function applies the formatting to a diarized text chunk.

    Parameters:
    ...........
    chunk: str
        Diarized text chunk.

    Returns:
    ...........
    dict: Input data for the API call.

    ------------------------------------------------------------------------------------------------------
    """
    prompt = PROMPT_FORMAT.replace("{{ user_msg_1 }}", chunk)
    res = {
        "inputs": prompt,
        "parameters": {
            "max_new_tokens":2048, "top_p":0.5, "temperature":0.2, "stop":["</s>", "###"]
        }
    }
    return res


def preprocess_str(text):
    """
    ------------------------------------------------------------------------------------------------------

    This function preprocesses a string by removing punctuation and multiple spaces.

    Parameters:
    ...........
    text: str
        Input string.

    Returns:
    ...........
    str: Preprocessed string.

    ------------------------------------------------------------------------------------------------------
    """
    text = text.replace("\n", " ").replace("\\", "")
    # remove punctuation - except for "'"
    text = "".join([c for c in text if c.isalnum() or c in [" ", "'", "<", ">", ":"]])
    # remove multiple spaces
    text = " ".join(text.split())
    text = text.lower().strip()
    return text


def extract_predicted_speakers(prompts, results, translate_json):
    """
    ------------------------------------------------------------------------------------------------------

    This function extracts the predicted speakers from the outputs of the speaker diarization model.

    Parameters:
    ...........
    prompts: dict
        Dictionary of diarized text chunks.
        key: chunk index, value: diarized text chunk.
    results: dict
        Dictionary of model outputs.
        key: chunk index, value: model output.
    translate_json: dict
        Dictionary to translate '1' and '2' to the original speaker labels.

    Returns:
    ...........
    speakers_pred: list
        List of predicted speaker labels for the entire transcription.

    ------------------------------------------------------------------------------------------------------
    """
    speakers_pred = []

    for idx in sorted(results.keys()):
        input = prompts[idx]
        output = results[idx].split("### Answer\n")[1].strip()

        input = preprocess_str(input)
        output = preprocess_str(output)

        # extract text and speaker
        words_in, speakers_in = extract_text_and_spk(input)
        words_out, speakers_out = extract_text_and_spk(output)

        if len(output) > 1.5 * len(input):
            # if the output is significantly longer than the input,
            # it is likely the model has hallucinated some words
            # thus, we will not apply the correction
            speakers_pred += [translate_json[x] for x in speakers_in.split()]
            continue

        speakers_out2 = transcript_preserving_speaker_transfer(
            words_out, speakers_out, words_in, speakers_in
        )

        speakers_pred += [translate_json[x] for x in speakers_out2.split()]

    return speakers_pred


def extract_prompts(transcript_json, asr):
    """
    ------------------------------------------------------------------------------------------------------

    This function extracts the prompts and the translation JSON from a transcription JSON.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.
    asr: str
        Automatic Speech Recognition (ASR)
         system used to transcribe the audio.

    Returns:
    ...........
    prompts: dict
        Dictionary of diarized text chunks.
        key: chunk index, value: diarized text chunk.
    translate_json: dict
        Dictionary to translate '1' and '2' to the original speaker labels.

    ------------------------------------------------------------------------------------------------------
    """
    if asr == "aws":
        speakers, words, translate_json = extract_transcription_aws(transcript_json)
    elif asr == "whisperx":
        speakers, words, translate_json = extract_transcription_whisperx(
            transcript_json
        )
    else:
        raise ValueError("ASR not supported")

    chunks = split_transcription(words, speakers)

    prompts = {idx: chunk for idx, chunk in enumerate(chunks)}

    return prompts, translate_json


class AWSProcessor:
    def __init__(self, endpoint_name, input_param):
        self.endpoint_name = endpoint_name
        self.input_param = input_param

    @exponential_backoff_decorator(max_retries=3, base_delay=90)
    def process_chunk(self, args):
        """
        ------------------------------------------------------------------------------------------------------
        This function processes a diarized text chunk using the SageMaker endpoint;
        it is used for parallel processing.

        Parameters:
        ...........
        args: tuple
            Tuple of arguments.
            idx: int, chunk index.
            prompt: str, diarized text chunk.

        Returns:
        ...........
        tuple: Tuple of chunk index and model output.

        ------------------------------------------------------------------------------------------------------
        """
        idx, prompt = args

        # initialize boto3 client
        if self.input_param['access_key'] and self.input_param['secret_key']:
            client = boto3.client('sagemaker-runtime', region_name=self.input_param['region'], 
                                        aws_access_key_id=self.input_param['access_key'], 
                                        aws_secret_access_key=self.input_param['secret_key'])
        else:
            client = boto3.client('sagemaker-runtime', region_name=self.input_param['region'])

        input_data = apply_formatting(preprocess_str(prompt))

        # Convert input data to JSON string
        payload = json.dumps(input_data)

        # Invoke the endpoint
        response = client.invoke_endpoint(
            EndpointName=self.endpoint_name,
            Body=payload,
            ContentType='application/json'
        )

        # Parse the response
        result = json.loads(response['Body'].read().decode())[0]["generated_text"]

        return idx, result


def call_diarization(prompts, endpoint_name, input_param):
    """
    ------------------------------------------------------------------------------------------------------

    This function calls the speaker diarization model.

    Parameters:
    ...........
    prompts: dict
        Dictionary of diarized text chunks.
        key: chunk index, value: diarized text chunk.
    endpoint_name: str
        Name of the SageMaker endpoint.
    input_param: dict
        Additional arguments for the API call.

    Returns:
    ...........
    results: dict
        Dictionary of model outputs.
        key: chunk index, value: model output.

    ------------------------------------------------------------------------------------------------------
    """

    processor = AWSProcessor(endpoint_name, input_param)
    results = {}
    if input_param['parallel_processing'] == 1:
        with Pool(processes=len(prompts)) as pool:
            args = [(idx, prompts[idx]) for idx in sorted(prompts.keys())]
            results = pool.map(processor.process_chunk, args)
            results = dict(results)
    else:
        for idx in sorted(prompts.keys()):
            results[idx] = processor.process_chunk((idx, prompts[idx]))[1]

    return results


def correct_transcription(transcript_json, prompts, results, translate_json, asr):
    """
    ------------------------------------------------------------------------------------------------------

    This function performs completion parsing of the LLM diarization model outputs
     and reinserts the corrected speaker labels into the transcription JSON.

    Parameters:
    ...........
    transcript_json: dict
        JSON response object.
    prompts: dict
        Dictionary of diarized text chunks.
        key: chunk index, value: diarized text chunk.
    results: dict
        Dictionary of model outputs.
        key: chunk index, value: model output.
    translate_json: dict
        Dictionary to translate '1' and '2' to the original speaker labels.
    asr: str
        Automatic Speech Recognition (ASR)
         system used to transcribe the audio.

    Returns:
    ...........
    dict: JSON response object with corrected speaker diarization.

    ------------------------------------------------------------------------------------------------------
    """
    speakers_pred = extract_predicted_speakers(prompts, results, translate_json)

    if asr == "aws":
        transcript_json_corrected = modify_transcription_aws(
            transcript_json, speakers_pred
        )
    elif asr == "whisperx":
        transcript_json_corrected = modify_transcription_whisperx(
            transcript_json, speakers_pred
        )
    else:
        raise ValueError("ASR not supported")

    return transcript_json_corrected



import pandas as pd
import numpy as np

def transcribe_response_to_dataframe(response):
    """
    ------------------------------------------------------------------------------------------------------

    Transcribes(cloud:AWS) a response from a speech-to-text service into a pandas DataFrame.

    Parameters:
    ----------
    response : dict
        The response object containing the transcribed data.

    Returns:
    -------
    df : pandas DataFrame
        The transcribed data in a DataFrame.
    speakers: int
        The number of speakers detected in the transcription.

    ------------------------------------------------------------------------------------------------------
    """
    speakers = 0
    df = pd.DataFrame()

    if 'results' in response:
        if 'speaker_labels' in response['results']:

            if 'speakers' in response['results']['speaker_labels']:
                speakers = response['results']['speaker_labels']["speakers"]

            if 'items' in response['results']:
                items = response['results']["items"]
                df = pd.DataFrame(items)

                df["confidence"] = df["alternatives"].apply(lambda x: x[0]["confidence"])
                df["content"] = df["alternatives"].apply(lambda x: x[0]["content"])
                df["confidence"] = df["confidence"].astype(float)

                df = df[df["confidence"] > 0].reset_index(drop=True)
                df = df[["start_time", "end_time", "confidence", "speaker_label", "content"]]
                
    return df, speakers

def extract_data(segment_info):
    """
    ------------------------------------------------------------------------------------------------------

    extract data from word_info

    Parameters:
    ----------
    segment_info : object
        The phrase level transcribed data.

    Returns:
    -------
    df : pandas series
        The phrase level transcribed data in a pandas series.

    ------------------------------------------------------------------------------------------------------
    """
    words = segment_info.get("words", None)

    starts = [word.get("start", np.nan) for word in words]
    ends = [word.get("end", np.nan) for word in words]
    phrases = [word.get("word", "") for word in words]
    scores = [word.get("score", 0) for word in words]
    speakers = [segment_info.get("speaker", "no_speaker") for _ in words]

    return pd.DataFrame({"start": starts, "end": ends, "phrase": phrases, "score": scores, "speaker": speakers})

def whisperx_to_dataframe(json_response):
    """
    ------------------------------------------------------------------------------------------------------

    Transcribes(local:whisperx) a json response into a pandas DataFrame.

    Parameters:
    ----------
    json_response : dict
        The response object containing the transcribed data.

    Returns:
    -------
    df : pandas DataFrame
        The transcribed data in a DataFrame.
    speakers: int
        The number of speakers detected in the transcription.

    ------------------------------------------------------------------------------------------------------
    """
    df = pd.DataFrame(columns=["start_time", "end_time", "content", "confidence", "speaker_label"])
    if 'segments' in json_response:
        
        segment_infos = json_response["segments"]
        df = pd.DataFrame(segment_infos).apply(extract_data, axis=1)
        df = pd.concat(df.tolist(), ignore_index=True)

        df = df[df["score"] > 0].reset_index(drop=True)
        df = df.dropna(subset=["start", "end"]).reset_index(drop=True)
        
        df = df[df["speaker"] != "no_speaker"].reset_index(drop=True)
        df = df.rename(columns={"start": "start_time", "end": "end_time", "score": "confidence", "speaker": "speaker_label", "phrase": "content"})

    speakers = df['speaker_label'].nunique()
    return df, speakers

def vosk_to_dataframe(json_response):
    """
    ------------------------------------------------------------------------------------------------------

    Transcribes(local:vosk) a json response into a pandas DataFrame.

    Parameters:
    ----------
    json_response : dict
        The response object

    Returns:
    -------
    df : pandas DataFrame
        The transcribed data in a DataFrame.

    ------------------------------------------------------------------------------------------------------
    """
    df = pd.DataFrame(columns=["start_time", "end_time", "content", "confidence", "speaker_label"])
        
    df = pd.DataFrame(json_response)

    df = df[df["conf"] > 0].reset_index(drop=True)
    df = df.dropna(subset=["start", "end"]).reset_index(drop=True)
    
    df = df.rename(columns={"start": "start_time", "end": "end_time", "conf": "confidence", "word": "content"})
    df['speaker_label'] = 'speaker0'

    return df

def volume_normalization(audio_signal, target_dBFS):
    """
    ------------------------------------------------------------------------------------------------------
    
    Normalizes the volume of the audio signal to the target dBFS.
    
    Parameters:
    ...........
    audio_signal : pydub.AudioSegment
        input audio signal
    target_dBFS : float
        target dBFS
        
    Returns:
    ...........
    pydub.AudioSegment
        normalized audio signal

    ------------------------------------------------------------------------------------------------------
    """

    headroom = -audio_signal.max_dBFS
    gain_adjustment = target_dBFS - audio_signal.dBFS

    if gain_adjustment > headroom:
        gain_adjustment = headroom

    audio_signal = audio_signal.apply_gain(gain_adjustment)
    return audio_signal
